// file      : xsd/cxx/tree/serialization-source.cxx
// author    : Boris Kolpackov <boris@codesynthesis.com>
// copyright : Copyright (c) 2005-2007 Code Synthesis Tools CC
// license   : GNU GPL v2 + exceptions; see accompanying LICENSE file

#include <cxx/tree/serialization-source.hxx>

#include <xsd-frontend/semantic-graph.hxx>
#include <xsd-frontend/traversal.hxx>

namespace CXX
{
  namespace Tree
  {
    namespace
    {
      struct List : Traversal::List, protected virtual Context
      {
        List (Context& c)
            : Context (c)
        {
        }

        virtual Void
        traverse (Type& l)
        {
          String name (ename (l));
          String error_handler (xs_ns_name () + L"::error_handler");

          // If renamed name is empty then we do not need to generate
          // anything for this type.
          //
          if (renamed_type (l, name) && !name)
            return;

          // operator<< (xercesc::DOMElement)
          //
          os << "void" << endl
             << "operator<< (::xercesc::DOMElement& e, " <<
            "const " << name << "& i)"
             << "{"
             << "e << static_cast< const ::xsd::cxx::tree::list< " <<
            item_type_name (l.argumented ().type ()) << ", " <<
            char_type << " >& > (i);"
             << "}";

          // operator<< (xercesc::DOMAttr)
          //
          os << "void" << endl
             << "operator<< (::xercesc::DOMAttr& a, " <<
            "const " << name << "& i)"
             << "{"
             << "a << static_cast< const ::xsd::cxx::tree::list< " <<
            item_type_name (l.argumented ().type ()) << ", "
             << char_type << " >& > (i);"
             << "}";

          // operator<< (list_stream)
          //
          os << "void" << endl
             << "operator<< (::xsd::cxx::tree::list_stream< " <<
            char_type << " >& l," << endl
             << "const " << name << "& i)"
             << "{"
             << "l << static_cast< const ::xsd::cxx::tree::list< " <<
            item_type_name (l.argumented ().type ()) << ", " <<
            char_type << " >& > (i);"
             << "}";

          // Register with type factory map.
          //
          if (options.value<CLI::generate_polymorphic> () &&
              !l.context ().count ("anonymous"))
          {
            // Note that we are using the original type name.
            //
            os << "static" << endl
               << "::xsd::cxx::tree::type_serializer_initializer< 0, " <<
              char_type << ", " << ename (l) << " >" << endl
               << "_xsd_" << ename (l) << "_type_serializer_init (" << endl
               << L << "\"" << l.name () << "\"," << endl
               << L << "\"" << xml_ns_name (l) << "\");"
               << endl
               << endl;
          }
        }

      private:
        String
        item_type_name (SemanticGraph::Type& t)
        {
          std::wostringstream o;

          MemberTypeName type (*this, o);
          type.dispatch (t);

          return o.str ();
        }
      };


      struct Union : Traversal::Union, protected virtual Context
      {
        Union (Context& c)
            : Context (c)
        {
        }

        virtual Void
        traverse (Type& u)
        {
          String name (ename (u));

          // If renamed name is empty then we do not need to generate
          // anything for this type.
          //
          if (renamed_type (u, name) && !name)
            return;

          String base (xs_ns_name () + L"::string");

          // operator<< (xercesc::DOMElement)
          //
          os << "void" << endl
             << "operator<< (::xercesc::DOMElement& e, " <<
            "const " << name << "& i)"
             << "{"
             << "e << static_cast< const " << base << "& > (i);"
             << "}";

          // operator<< (xercesc::DOMAttr)
          //
          os << "void" << endl
             << "operator<< (::xercesc::DOMAttr& a, " <<
            "const " << name << "& i)"
             << "{"
             << "a << static_cast< const " << base << "& > (i);"
             << "}";

          // operator<< (list_stream)
          //
          os << "void" << endl
             << "operator<< (::xsd::cxx::tree::list_stream< " <<
            char_type << " >& l," << endl
             << "const " << name << "& i)"
             << "{"
             << "l << static_cast< const " << base << "& > (i);"
             << "}";

          // Register with type factory map.
          //
          if (options.value<CLI::generate_polymorphic> () &&
              !u.context ().count ("anonymous"))
          {
            // Note that we are using the original type name.
            //
            os << "static" << endl
               << "::xsd::cxx::tree::type_serializer_initializer< 0, " <<
              char_type << ", " << ename (u) << " >" << endl
               << "_xsd_" << ename (u) << "_type_serializer_init (" << endl
               << L << "\"" << u.name () << "\"," << endl
               << L << "\"" << xml_ns_name (u) << "\");"
               << endl
               << endl;
          }
        }
      };


      struct Enumeration : Traversal::Enumeration, protected virtual Context
      {
        Enumeration (Context& c)
            : Context (c), base_ (c)
        {
          inherits_base_ >> base_;
        }

        virtual Void
        traverse (Type& e)
        {
          String name (ename (e));

          // If renamed name is empty then we do not need to generate
          // anything for this type.
          //
          if (renamed_type (e, name) && !name)
            return;

          // operator<< (xercesc::DOMElement)
          //
          os << "void" << endl
             << "operator<< (::xercesc::DOMElement& e, " <<
            "const " << name << "& i)"
             << "{"
             << "e << static_cast< const ";

          inherits (e, inherits_base_);

          os << "& > (i);"
             << "}";


          // operator<< (xercesc::DOMAttr)
          //
          os << "void" << endl
             << "operator<< (::xercesc::DOMAttr& a, " <<
            "const " << name << "& i)"
             << "{"
             << "a << static_cast< const ";

          inherits (e, inherits_base_);

          os << "& > (i);"
             << "}";


          // operator<< (list_stream)
          //
          os << "void" << endl
             << "operator<< (::xsd::cxx::tree::list_stream< " <<
            char_type << " >& l," << endl
             << "const " << name << "& i)"
             << "{"
             << "l << static_cast< const ";

          inherits (e, inherits_base_);

          os << "& > (i);"
             << "}";


          // Register with type factory map.
          //
          if (options.value<CLI::generate_polymorphic> () &&
              !e.context ().count ("anonymous"))
          {
            // Note that we are using the original type name.
            //
            os << "static" << endl
               << "::xsd::cxx::tree::type_serializer_initializer< 0, " <<
              char_type << ", " << ename (e) << " >" << endl
               << "_xsd_" << ename (e) << "_type_serializer_init (" << endl
               << L << "\"" << e.name () << "\"," << endl
               << L << "\"" << xml_ns_name (e) << "\");"
               << endl
               << endl;
          }
        }

      private:
        Traversal::Inherits inherits_base_;
        BaseTypeName base_;
      };

      struct Element : Traversal::Element, protected virtual Context
      {
        Element (Context& c, String const& scope_)
            : Context (c), scope (scope_)
        {
        }

        virtual Void
        traverse (Type& e)
        {
          if (skip (e))
            return;

          String const& name (ename (e));
          String ns (e.qualified () ? e.namespace_ ().name () : "");
          String type (scope + L"::" + etype (e));

          // Check if we need to handle xsi:type and substitution groups.
          // If this element's type is anonymous or mapped to a fundamental
          // C++ type then we don't need to do anything. Note that if the
          // type is anonymous then it can't be derived from which makes it
          // impossible to substitute or dynamically-type with xsi:type.
          //
          SemanticGraph::Type& t (e.type ());

          Boolean polymorphic (options.value<CLI::generate_polymorphic> () &&
                               !t.context ().count ("anonymous"));

          if (polymorphic)
          {
            Boolean fund (false);
            IsFundamentalType traverser (fund);
            traverser.dispatch (t);
            polymorphic = !fund;
          }

          os << "// " << name << endl
             << "//" << endl;

          // aCC cannot handle an inline call to type_serializer_map_instance.
          //
          if (polymorphic)
          {
            os << "{"
               << "::xsd::cxx::tree::type_serializer_map< " << char_type
               << " >& tsm (" << endl
               << "::xsd::cxx::tree::type_serializer_map_instance< 0, " <<
              char_type << " > ());"
               << endl;
          }

          if (max (e) != 1)
          {
            // sequence
            //
            os << "for (" << scope << "::" << econst_iterator (e) << endl
               << "b (i." << name << " ().begin ()), " <<
              "n (i." << name << " ().end ());" << endl
               << "b != n; ++b)"
               << "{";

            if (polymorphic)
            {
              os << "if (typeid (" << type << ") == typeid (*b))"
                 << "{"
                 << "::xercesc::DOMElement& s (" << endl
                 << "::xsd::cxx::xml::dom::create_element (" << endl
                 << L << "\"" << e.name () << "\"," << endl
                 << (ns ? L + L"\"" + ns + L"\"" + L",\n" : L"")
                 << "e));"
                 << endl
                 << "s << *b;"
                 << "}"
                 << "else" << endl
                 << "tsm.serialize (" << endl
                 << L << "\"" << e.name () << "\"," << endl
                 << L << "\"" << ns << "\"," << endl
                 << (e.global () ? "true" : "false") << ", " <<
                (e.qualified () ? "true" : "false") << ", e, *b);";
            }
            else
            {
              os << "::xercesc::DOMElement& s (" << endl
                 << "::xsd::cxx::xml::dom::create_element (" << endl
                 << L << "\"" << e.name () << "\"," << endl
                 << (ns ? L + L"\"" + ns + L"\"" + L",\n" : L"")
                 << "e));"
                 << endl
                 << "s << *b;";
            }


            os << "}";
          }
          else if (min (e) == 0)
          {
            // optional
            //
            os << "if (i." << name << " ())"
               << "{";

            if (polymorphic)
            {
              os << "const " << type << "& x (*i." << name << " ());"
                 << "if (typeid (" << type << ") == typeid (x))"
                 << "{"
                 << "::xercesc::DOMElement& s (" << endl
                 << "::xsd::cxx::xml::dom::create_element (" << endl
                 << L << "\"" << e.name () << "\"," << endl
                 << (ns ? L + L"\"" + ns + L"\"" + L",\n" : L"")
                 << "e));"
                 << endl
                 << "s << x;"
                 << "}"
                 << "else" << endl
                 << "tsm.serialize (" << endl
                 << L << "\"" << e.name () << "\"," << endl
                 << L << "\"" << ns << "\"," << endl
                 << (e.global () ? "true" : "false") << ", " <<
                (e.qualified () ? "true" : "false") << ", e, x);";
            }
            else
            {
              os << "::xercesc::DOMElement& s (" << endl
                 << "::xsd::cxx::xml::dom::create_element (" << endl
                 << L << "\"" << e.name () << "\"," << endl
                 << (ns ? L + L"\"" + ns + L"\"" + L",\n" : L"")
                 << "e));"
                 << endl
                 << "s << *i." << name << " ();";
            }

            os << "}";
          }
          else
          {
            // one
            //
            if (polymorphic)
            {
              os << "const " << type << "& x (i." << name << " ());"
                 << "if (typeid (" << type << ") == typeid (x))"
                 << "{"
                 << "::xercesc::DOMElement& s (" << endl
                 << "::xsd::cxx::xml::dom::create_element (" << endl
                 << L << "\"" << e.name () << "\"," << endl
                 << (ns ? L + L"\"" + ns + L"\"" + L",\n" : L"")
                 << "e));"
                 << endl
                 << "s << x;"
                 << "}"
                 << "else" << endl
                 << "tsm.serialize (" << endl
                 << L << "\"" << e.name () << "\"," << endl
                 << L << "\"" << ns << "\"," << endl
                 << (e.global () ? "true" : "false") << ", " <<
                (e.qualified () ? "true" : "false") << ", e, x);";
            }
            else
            {
              os << "{"
                 << "::xercesc::DOMElement& s (" << endl
                 << "::xsd::cxx::xml::dom::create_element (" << endl
                 << L << "\"" << e.name () << "\"," << endl
                 << (ns ? L + L"\"" + ns + L"\"" + L",\n" : L"")
                 << "e));"
                 << endl
                 << "s << i." << name << " ();"
                 << "}";
            }
          }

          if (polymorphic)
            os << "}";
        }

      private:
        String scope;
      };

      struct Any : Traversal::Any, protected virtual Context
      {
        Any (Context& c, String const& scope_)
            : Context (c), scope (scope_)
        {
        }

        virtual Void
        traverse (Type& a)
        {
          String const& name (ename (a));

          os << "// " << name << endl
             << "//" << endl;

          if (max (a) != 1)
          {
            // sequence
            //
            os << "for (" << scope << "::" << econst_iterator (a) << endl
               << "b (i." << name << " ().begin ()), " <<
              "n (i." << name << " ().end ());" << endl
               << "b != n; ++b)"
               << "{"
               << "e.appendChild (" << endl
               << "e.getOwnerDocument ()->importNode (" << endl
               << "const_cast< ::xercesc::DOMElement* > (&(*b)), true));"
               << "}";
          }
          else if (min (a) == 0)
          {
            // optional
            //
            os << "if (i." << name << " ())"
               << "{"
               << "e.appendChild (" << endl
               << "e.getOwnerDocument ()->importNode (" << endl
               << "const_cast< ::xercesc::DOMElement* > (&(*i." <<
              name << " ())), true));"
               << "}";
          }
          else
          {
            // one
            //
            os << "e.appendChild (" << endl
               << "e.getOwnerDocument ()->importNode (" << endl
               << "const_cast< ::xercesc::DOMElement* > (&(i." <<
              name << " ())), true));"
               << endl;
          }
        }

      private:
        String scope;
      };

      struct Attribute : Traversal::Attribute, protected virtual Context
      {
        Attribute (Context& c)
            : Context (c)
        {
        }

        virtual Void
        traverse (Type& a)
        {
          String const& name (ename (a));
          String ns (a.qualified () ? a.namespace_ ().name () : "");

          os << "// " << name << endl
             << "//" << endl;

          if (a.optional () && !a.default_ ())
          {
            os << "if (i." << name << " ())"
               << "{"
               << "::xercesc::DOMAttr& a (" << endl
               << "::xsd::cxx::xml::dom::create_attribute (" << endl
               << L << "\"" << a.name () << "\"," << endl
               << (ns ? L + L"\"" + ns + L"\"" + L",\n" : L"")
               << "e));"
               << endl
               << "a << *i." << name << " ();"
               << "}";
          }
          else
          {
            os << "{"
               << "::xercesc::DOMAttr& a (" << endl
               << "::xsd::cxx::xml::dom::create_attribute (" << endl
               << L << "\"" << a.name () << "\"," << endl
               << (ns ? L + L"\"" + ns + L"\"" + L",\n" : L"")
               << "e));"
               << endl
               << "a << i." << name << " ();"
               << "}";
          }
        }
      };

      struct AnyAttribute : Traversal::AnyAttribute,
                            protected virtual Context
      {
        AnyAttribute (Context& c, String const& scope_)
            : Context (c), scope (scope_)
        {
        }

        virtual Void
        traverse (Type& a)
        {
          String const& name (ename (a));

          os << "// " << name << endl
             << "//" << endl;

          os << "for (" << scope << "::" << econst_iterator (a) << endl
             << "b (i." << name << " ().begin ()), " <<
            "n (i." << name << " ().end ());" << endl
             << "b != n; ++b)"
             << "{"
             << "::xercesc::DOMAttr* a (" << endl
             << "static_cast< ::xercesc::DOMAttr* > (" << endl
             << "e.getOwnerDocument ()->importNode (" << endl
             << "const_cast< ::xercesc::DOMAttr* > (&(*b)), true)));"
             << endl
             << "if (a->getLocalName () == 0)" << endl
             << "e.setAttributeNode (a);"
             << "else" << endl
             << "e.setAttributeNodeNS (a);"
             << "}";
        }

      private:
        String scope;
      };

      struct Complex : Traversal::Complex, protected virtual Context
      {
        Complex (Context& c)
            : Context (c), base_ (c)
        {
          inherits_ >> base_;
        }

        virtual Void
        traverse (Type& c)
        {
          String name (ename (c));

          // If renamed name is empty then we do not need to generate
          // anything for this type.
          //
          if (renamed_type (c, name) && !name)
            return;

          // operator<< (xercesc::DOMElement)
          //
          os << "void" << endl
             << "operator<< (::xercesc::DOMElement& e, " <<
            "const " << name << "& i)"
             << "{";

          if (c.inherits_p ())
          {
            os << "e << static_cast< const ";
            inherits (c, inherits_);
            os << "& > (i);"
               << endl;
          }
          else
            os << "e << static_cast< const " << any_type << "& > (i);"
               << endl;

          // Serialize anyAttribute content first so that is gets
          // overriden by schema-defined attributes.
          //
          if (options.value<CLI::generate_wildcard> ())
          {
            AnyAttribute any_attribute (*this, name);
            Traversal::Names names (any_attribute);

            Complex::names (c, names);
          }

          {
            Traversal::Names names;
            Any any (*this, name);
            Element element (*this, name);
            Attribute attribute (*this);

            names >> element;
            names >> attribute;

            if (options.value<CLI::generate_wildcard> ())
              names >> any;

            Complex::names (c, names);
          }

          os << "}";


          Boolean simple (true);
          {
            IsSimpleType t (simple);
            t.dispatch (c);
          }

          if (simple)
          {
            Boolean hb (c.inherits_p ());

            // operator<< (xercesc::DOMAttr)
            //
            os << "void" << endl
               << "operator<< (::xercesc::DOMAttr&" <<
              (hb ? " a" : "") << "," <<
              "const " << name << "&" << (hb ? " i" : "") << ")"
               << "{";

            if (hb)
            {
              os << "a << static_cast< const ";

              inherits (c, inherits_);

              os << "& > (i);";
            }

            os << "}";


            // operator<< (list_stream)
            //
            os << "void" << endl
               << "operator<< (::xsd::cxx::tree::list_stream< " <<
              char_type << " >&" << (hb ? " l" : "") << "," << endl
               << "const " << name << "&" << (hb ? " i" : "") << ")"
               << "{";

            if (hb)
            {
              os << "l << static_cast< const ";

              inherits (c, inherits_);

              os << "& > (i);";
            }

            os << "}";
          }

          // Register with type factory map.
          //
          if (options.value<CLI::generate_polymorphic> () &&
              !c.context ().count ("anonymous"))
          {
            // Note that we are using the original type name.
            //
            os << "static" << endl
               << "::xsd::cxx::tree::type_serializer_initializer< 0, " <<
              char_type << ", " << ename (c) << " >" << endl
               << "_xsd_" << ename (c) << "_type_serializer_init (" << endl
               << L << "\"" << c.name () << "\"," << endl
               << L << "\"" << xml_ns_name (c) << "\");"
               << endl
               << endl;
          }
        }

      private:
        Traversal::Inherits inherits_;
        BaseTypeName base_;
      };


      // Generate substitution group map entry.
      //
      struct GlobalMember : Traversal::Element,
                            GlobalElementBase,
                            protected virtual Context
      {
        GlobalMember (Context& c)
            : GlobalElementBase (c), Context (c), type_name_ (c)
        {
          belongs_ >> type_name_;
        }

        virtual Void
        traverse (Type& e)
        {
          if (!generate_p (e))
            return;

          if (e.substitutes_p () &&
	      options.value<CLI::generate_polymorphic> ())
          {
            Type& r (e.substitutes ().root ());

            String const& name (ename (e));

            os << "static" << endl
               << "::xsd::cxx::tree::type_serializer_initializer< 0, " <<
              char_type << ", ";

            belongs (e, belongs_);

            os << " >" << endl
               << "_xsd_" << name << "_element_name_init (" << endl
               << L << "\"" << r.name () << "\"," << endl
               << L << "\"" << r.namespace_ ().name () << "\"," << endl
               << L << "\"" << e.name () << "\"," << endl
               << L << "\"" << e.namespace_ ().name () << "\");"
               << endl
               << endl;
          }
        }

      private:
        Traversal::Belongs belongs_;
        MemberTypeName type_name_;

      };

      struct Serializer: Traversal::Element,
                         GlobalElementBase,
                         protected virtual Context
      {
        Serializer (Context& c)
            : GlobalElementBase (c), Context (c)
        {
        }

        virtual Void
        traverse (Type& e)
        {
          if (!doc_root_p (e))
            return;

          String const& name (ename (e));
          String ns (e.namespace_ ().name ());
          String error_handler (xs_ns_name () + L"::error_handler");
          String namespace_infomap (xs_ns_name () + L"::namespace_infomap");

          // Note that I am using fq-name in function calls because g++ gets
          // confused if the name is 'type'. (see tests/schema/anonymous)
          //

          Boolean fund (false);
          {
            IsFundamentalType t (fund);
            t.dispatch (e.type ());
          }

          // Check if we need to handle xsi:type and substitution groups.
          // If this element's type is anonymous or mapped to a fundamental
          // C++ type then we don't need to do anything.
          //
          Boolean polymorphic (!fund &&
                               options.value<CLI::generate_polymorphic> () &&
                               !e.type ().context ().count ("anonymous"));

          // To std::ostream.
          //
          os << "void" << endl
             << name << " (::std::ostream& o," << endl
             << "const " << type_name (e) << "& s," << endl
             << "const " << namespace_infomap << "& m," << endl
             << "const " << string_type << "& e," << endl
             << flags_type << " f)"
             << "{"
             << "::xsd::cxx::xml::auto_initializer i (" << endl
             << "(f & " << flags_type << "::dont_initialize) == 0);"
             << endl
             << "::xsd::cxx::xml::dom::auto_ptr< " <<
            "::xercesc::DOMDocument > d (" << endl
             << fq_name (e) << " (s, m, f));"
             << endl
             << "::xsd::cxx::tree::error_handler< " << char_type << " > h;"
             << endl
             << "::xsd::cxx::xml::dom::ostream_format_target t (o);"
             << "if (!::xsd::cxx::xml::dom::serialize (t, *d, e, h, f))"
             << "{"
             << "h.throw_if_failed< ::xsd::cxx::tree::serialization< " <<
            char_type << " > > ();"
             << "}"
             << "}";

          os << "void" << endl
             << name << " (::std::ostream& o," << endl
             << "const " << type_name (e) << "& s," << endl
             << "const " << namespace_infomap << "& m," << endl
             << error_handler << "& h," << endl
             << "const " << string_type << "& e," << endl
             << flags_type << " f)"
             << "{"
             << "::xsd::cxx::xml::auto_initializer i (" << endl
             << "(f & " << flags_type << "::dont_initialize) == 0);"
             << endl
             << "::xsd::cxx::xml::dom::auto_ptr< " <<
            "::xercesc::DOMDocument > d (" << endl
             << fq_name (e) << " (s, m, f));"
             << "::xsd::cxx::xml::dom::ostream_format_target t (o);"
             << "if (!::xsd::cxx::xml::dom::serialize (t, *d, e, h, f))"
             << "{"
             << "throw ::xsd::cxx::tree::serialization< " <<
            char_type << " > ();"
             << "}"
             << "}";

          os << "void" << endl
             << name << " (::std::ostream& o," << endl
             << "const " << type_name (e) << "& s," << endl
             << "const " << namespace_infomap << "& m," << endl
             << "::xercesc::DOMErrorHandler& h," << endl
             << "const " << string_type << "& e," << endl
             << flags_type << " f)"
             << "{"
             << "::xsd::cxx::xml::dom::auto_ptr< " <<
            "::xercesc::DOMDocument > d (" << endl
             << fq_name (e) << " (s, m, f));"
             << "::xsd::cxx::xml::dom::ostream_format_target t (o);"
             << "if (!::xsd::cxx::xml::dom::serialize (t, *d, e, h, f))"
             << "{"
             << "throw ::xsd::cxx::tree::serialization< " <<
            char_type << " > ();"
             << "}"
             << "}";

          // To XMLFormatTarget.
          //
          os << "void" << endl
             << name << " (::xercesc::XMLFormatTarget& t," << endl
             << "const " << type_name (e) << "& s," << endl
             << "const " << namespace_infomap << "& m," << endl
             << "const " << string_type << "& e," << endl
             << flags_type << " f)"
             << "{"
             << "::xsd::cxx::xml::dom::auto_ptr< " <<
            "::xercesc::DOMDocument > d (" << endl
             << fq_name (e) << " (s, m, f));"
             << endl
             << "::xsd::cxx::tree::error_handler< " << char_type << " > h;"
             << endl
             << "if (!::xsd::cxx::xml::dom::serialize (t, *d, e, h, f))"
             << "{"
             << "h.throw_if_failed< ::xsd::cxx::tree::serialization< " <<
            char_type << " > > ();"
             << "}"
             << "}";

          os << "void" << endl
             << name << " (::xercesc::XMLFormatTarget& t," << endl
             << "const " << type_name (e) << "& s," << endl
             << "const " << namespace_infomap << "& m," << endl
             << error_handler << "& h," << endl
             << "const " << string_type << "& e," << endl
             << flags_type << " f)"
             << "{"
             << "::xsd::cxx::xml::dom::auto_ptr< " <<
            "::xercesc::DOMDocument > d (" << endl
             << fq_name (e) << " (s, m, f));"
             << "if (!::xsd::cxx::xml::dom::serialize (t, *d, e, h, f))"
             << "{"
             << "throw ::xsd::cxx::tree::serialization< " <<
            char_type << " > ();"
             << "}"
             << "}";

          os << "void" << endl
             << name << " (::xercesc::XMLFormatTarget& t," << endl
             << "const " << type_name (e) << "& s," << endl
             << "const " << namespace_infomap << "& m," << endl
             << "::xercesc::DOMErrorHandler& h," << endl
             << "const " << string_type << "& e," << endl
             << flags_type << " f)"
             << "{"
             << "::xsd::cxx::xml::dom::auto_ptr< " <<
            "::xercesc::DOMDocument > d (" << endl
             << fq_name (e) << " (s, m, f));"
             << "if (!::xsd::cxx::xml::dom::serialize (t, *d, e, h, f))"
             << "{"
             << "throw ::xsd::cxx::tree::serialization< " <<
            char_type << " > ();"
             << "}"
             << "}";

          // To an existing DOM instance.
          //
          os << "void" << endl
             << name << " (::xercesc::DOMDocument& d," << endl
             << "const " << type_name (e) << "& s," << endl
             << flags_type << ")"
             << "{"
             << "::xercesc::DOMElement& e (*d.getDocumentElement ());"
             << "const " << qname_type << " n (" << endl
             << "::xsd::cxx::xml::dom::name< " << char_type << " > (e));"
             << endl;

          if (polymorphic)
          {
            os << "if (typeid (" << type_name (e) << ") == typeid (s))"
               << "{";
          }

          os << "if (n.name () == " << L << "\"" << e.name () << "\""
             << " &&" << endl
             << "n.namespace_ () == " << L << "\"" << ns << "\")"
             << "{"
             << "e << s;"
             << "}"
             << "else"
             << "{"
             << "throw ::xsd::cxx::tree::unexpected_element < " <<
            char_type << " > (" << endl
             << "n.name ()," << endl
             << "n.namespace_ ()," << endl
             << L << "\"" << e.name () << "\"," << endl
             << L << "\"" << ns << "\");"
             << "}";

          if (polymorphic)
          {
            // aCC cannot handle an inline call to
            // type_serializer_map_instance.
            //
            os << "}"
               << "else"
               << "{"
               << "::xsd::cxx::tree::type_serializer_map< " << char_type
               << " >& tsm (" << endl
               << "::xsd::cxx::tree::type_serializer_map_instance< 0, " <<
              char_type << " > ());"
               << endl
               << "tsm.serialize (" << endl
               << L << "\"" << e.name () << "\"," << endl
               << L << "\"" << e.namespace_().name () << "\"," << endl
               << "e, n, s);"
               << "}";
          }

          os << "}";


          // To a new DOM instance.
          //
          os << "::xsd::cxx::xml::dom::auto_ptr< " <<
            "::xercesc::DOMDocument >" << endl
             << name << " (const " << type_name (e) << "& s," << endl
             << "const " << namespace_infomap << "& m," << endl
             << flags_type << " f)"
             << "{"
             << "try"
             << "{";

          if (polymorphic)
          {
            // aCC cannot handle an inline call to
            // type_serializer_map_instance.
            //
            os << "::xsd::cxx::xml::dom::auto_ptr< " <<
              "::xercesc::DOMDocument > d;"
               << endl
               << "if (typeid (" << type_name (e) << ") == typeid (s))"
               << "{"
               << "d = ::xsd::cxx::xml::dom::serialize< " <<
              char_type << " > (" << endl
               << L << "\"" << e.name () << "\"," << endl
               << L << "\"" << ns << "\"," << endl
               << "m, f);"
               << "}"
               << "else"
               << "{"
               << "::xsd::cxx::tree::type_serializer_map< " << char_type
               << " >& tsm (" << endl
               << "::xsd::cxx::tree::type_serializer_map_instance< 0, " <<
              char_type << " > ());"
               << endl
               << "d = tsm.serialize (" << endl
               << L << "\"" << e.name () << "\"," << endl
               << L << "\"" << e.namespace_().name () << "\"," << endl
               << "m, s, f);"
               << "}";
          }
          else
          {
            os << "::xsd::cxx::xml::dom::auto_ptr< " <<
              "::xercesc::DOMDocument > d (" << endl
               << "::xsd::cxx::xml::dom::serialize< " <<
              char_type << " > (" << endl
               << L << "\"" << e.name () << "\"," << endl
               << L << "\"" << ns << "\"," << endl
               << "m, f));"
               << endl;
          }

          os << fq_name (e) << " (*d, s, f);"
             << "return d;"
             << "}"
             << "catch (const ::xsd::cxx::xml::dom::mapping< " << char_type << " >& e)"
             << "{"
             << "throw ::xsd::cxx::tree::no_namespace_mapping< " << char_type << " > (e.name ());"
             << "}"
             << "catch (const ::xsd::cxx::xml::dom::xsi_already_in_use&)"
             << "{"
             << "throw ::xsd::cxx::tree::xsi_already_in_use< " <<
	     char_type << " > ();"
             << "}"
             << "}";

        }

      private:
        String
        type_name (Type& e)
        {
          std::wostringstream o;

          MemberTypeName type (*this, o);
          type.dispatch (e.type ());

          return o.str ();
        }
      };
    }

    Void
    generate_serialization_source (Context& ctx,
                                   UnsignedLong first,
                                   UnsignedLong last)
    {
      ctx.os << "#include <ostream>" << endl
             << "#include <xsd/cxx/xml/dom/serialization-source.hxx>" << endl
             << "#include <xsd/cxx/tree/error-handler.hxx>" << endl
             << endl;

      if (ctx.options.value<CLI::generate_polymorphic> ())
      {
        ctx.os << "#include <xsd/cxx/tree/type-serializer-map.hxx>" << endl
               << endl;

        ctx.os << "namespace _xsd"
               << "{"
               << "static" << endl
               << "::xsd::cxx::tree::type_serializer_plate< 0, " <<
          ctx.char_type << " >" << endl
               << "type_serializer_plate_init;"
               << "}";
      }

      Traversal::Schema schema;

      Traversal::Sources sources;
      Traversal::Names names_ns, names;

      Namespace ns (ctx, first, last);

      List list (ctx);
      Union union_ (ctx);
      Complex complex (ctx);
      Enumeration enumeration (ctx);
      GlobalMember member (ctx);
      Serializer serializer (ctx);


      schema >> sources >> schema;
      schema >> names_ns >> ns >> names;

      names >> list;
      names >> union_;
      names >> complex;
      names >> enumeration;
      names >> member;
      names >> serializer;


      schema.dispatch (ctx.schema_root);
    }
  }
}
